import jdatetime
import datetime
import time
import re
from django.db import models
from django.utils.translation import ugettext as _
from django.utils.encoding import smart_text, force_text, smart_str
from django_jalali import forms
from django import forms as mainforms
from django.utils.functional import curry
from django.core import exceptions
import django
from distutils.version import StrictVersion

ansi_date_re = re.compile(r'^\d{4}-\d{1,2}-\d{1,2}$')
class jManager(models.Manager):
    """we need to rewrite this class to handle year filter"""
    def filter(self, *args, **kwargs):
        """if filter is year we divide to __gte and __lte"""
        new_kwargs = {}
        for k in kwargs:
            if '__year' in k :
                filed_name = k.split('__year')
                first_year = jdatetime.datetime(int(kwargs[k]),1 ,1)
                new_kwargs['%s__gte'%filed_name[0]] = jdatetime.datetime(int(kwargs[k]),1 ,1)
                last_day = 29
                if first_year.isleap() :
                    last_day = 30
                new_kwargs['%s__lte'%filed_name[0]] = jdatetime.datetime(int(kwargs[k]),12, last_day, 23, 59, 59)
            else :
                new_kwargs[k] = kwargs[k]
        return models.Manager.filter(self, *args, **new_kwargs )



class jDateField(models.Field):
    description = _("Date (without time)")
    __metaclass__ = models.SubfieldBase
    empty_strings_allowed = False
    default_error_messages = {
        'invalid': _('Enter a valid date in YYYY-MM-DD format.'),
        'invalid_date': _('Invalid date: %s'),
    }

    def __init__(self, verbose_name=None, name=None, auto_now=False, auto_now_add=False, **kwargs):

        self.auto_now, self.auto_now_add = auto_now, auto_now_add
        #HACKs : auto_now_add/auto_now should be done as a default or a pre_save.
        if auto_now or auto_now_add:
            kwargs['editable'] = False
            kwargs['blank'] = True
        models.Field.__init__(self, verbose_name, name, **kwargs)

    def get_internal_type(self):
        return "DateField"

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            return jdatetime.date.fromgregorian(date=value.date())
        if isinstance(value, datetime.date):
            return jdatetime.date.fromgregorian(date=value)

        if isinstance(value, jdatetime.datetime):
            return value.date()
        if isinstance(value, jdatetime.date):
            return value

        if not ansi_date_re.search(value):
            raise exceptions.ValidationError(self.error_messages['invalid'])
        # Now that we have the date string in YYYY-MM-DD format, check to make
        # sure it's a valid date.
        # We could use time.strptime here and catch errors, but datetime.date
        # produces much friendlier error messages.
        year, month, day = map(int, value.split('-'))
        try:
            if year > 1500 :
                return jdatetime.date.fromgregorian(date=datetime.date(year, month, day))
            else:
                return jdatetime.date(year,month,day)
        except ValueError as e:
            msg = self.error_messages['invalid_date'] % _(str(e))
            raise exceptions.ValidationError(msg)

    def pre_save(self, model_instance, add):
        if self.auto_now or (self.auto_now_add and add):
            value = jdatetime.date.today()
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super(jDateField, self).pre_save(model_instance, add)

    def contribute_to_class(self, cls, name):
        super(jDateField,self).contribute_to_class(cls, name)
        if not self.null:
            setattr(cls, 'get_next_by_%s' % self.name,
                curry(cls._get_next_or_previous_by_FIELD, field=self, is_next=True))
            setattr(cls, 'get_previous_by_%s' % self.name,
                curry(cls._get_next_or_previous_by_FIELD, field=self, is_next=False))

    def get_prep_lookup(self, lookup_type, value):
        """this class dosn't work in month and day searh !"""
        # For "__month", "__day", and "__week_day" lookups, convert the value
        # to an int so the database backend always sees a consistent type.

        if lookup_type in ('exact', 'gt', 'gte', 'lt', 'lte'):
            prep  = self.get_prep_value(value)
            if type(prep) == datetime.datetime or type(prep) == datetime.date:
                return prep
            return prep.togregorian()

        elif lookup_type in ('range', 'in'):
            return [self.get_prep_value(v) for v in value]
        elif lookup_type == 'year':
            #this else never happen !
            try:
                return int(value)
            except ValueError:
                raise ValueError("The __year lookup type requires an integer argument")

        if lookup_type in ('month', 'day', 'week_day'):
            raise ValueError("jDateField dosn't work with month, day and week_day !")

        return super(jDateField, self).get_prep_lookup(lookup_type, value)

    def get_prep_value(self, value):
        return self.to_python(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        # Casts dates into the format expected by the backend
        if not prepared:
            value = self.get_prep_value(value)

        if isinstance(value, jdatetime.datetime):
            value = value.togregorian().date()
        if isinstance(value, jdatetime.date):
            value = value.togregorian()

        if StrictVersion(django.get_version()) >= StrictVersion('1.9'):
            return connection.ops.adapt_datefield_value(value)
        else :
            return connection.ops.value_to_db_date(value)

    def value_to_string(self, obj):
        val = self._get_val_from_obj(obj)
        if val is None:
            data = ''
        else:
            data = "te" #datetime_safe.new_date(val).strftime("%Y-%m-%d")
        return data

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.jDateField}
        defaults.update(kwargs)
        return super(jDateField, self).formfield(**defaults)


class jDateTimeField(jDateField):
    default_error_messages = {
        'invalid': _(u'Enter a valid date/time in YYYY-MM-DD HH:MM[:ss[.uuuuuu]] format.'),
    }
    description = _("Date (with time)")

    def get_internal_type(self):
        return "DateTimeField"

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            try :
                if value.year < 1700 :
                    return jdatetime.datetime(value.year, value.month, value.day, value.hour, value.minute, value.second, value.microsecond, value.tzinfo)
                else :
                    return jdatetime.datetime.fromgregorian(datetime=value)
            except ValueError :
                raise exceptions.ValidationError(self.error_messages['invalid'])
        if isinstance(value, datetime.date):
            try :
                if value.year < 1700 :
                    return jdatetime.datetime(value.year, value.month, value.day)
                else :
                    return jdatetime.datetime.fromgregorian(date=value)
            except ValueError :
                raise exceptions.ValidationError(self.error_messages['invalid'])
        if isinstance(value, jdatetime.datetime):
            return value
        if isinstance(value, jdatetime.date):
            try :
                d = jdatetime.datetime(value.year, value.month, value.day)
            except ValueError :
                raise exceptions.ValidationError(self.error_messages['invalid'])
            return d


        # Attempt to parse a datetime:
        value = smart_str(value)
        # split usecs, because they are not recognized by strptime.
        if '.' in value:
            try:
                value, usecs = value.split('.')
                usecs = int(usecs)
            except ValueError:
                raise exceptions.ValidationError(self.error_messages['invalid'])
        else:
            usecs = 0
        kwargs = {'microsecond': usecs}
        try: # Seconds are optional, so try converting seconds first.
            t = time.strptime(value, '%Y-%m-%d %H:%M:%S')
            if t.tm_year > 1700 :
                return datetime.datetime(*time.strptime(value, '%Y-%m-%d %H:%M:%S')[:6],
                                     **kwargs)
            else :
                return jdatetime.datetime(*time.strptime(value, '%Y-%m-%d %H:%M:%S')[:6],
                                    **kwargs)

        except ValueError:
            try: # Try without seconds.
                t = time.strptime(value, '%Y-%m-%d %H:%M')
                if t.tm_year > 1700:
                    return  datetime.datetime(*time.strptime(value, '%Y-%m-%d %H:%M')[:5],
                                     **kwargs)
                else :
                    return  jdatetime.datetime(*time.strptime(value, '%Y-%m-%d %H:%M')[:5],
                                     **kwargs)

            except ValueError: # Try without hour/minutes/seconds.
                try:
                    t = time.strptime(value, '%Y-%m-%d')[:3]
                    if t[0] > 1700 :
                        return datetime.datetime(*time.strptime(value, '%Y-%m-%d')[:3],
                                             **kwargs)
                    else :
                        return jdatetime.datetime(*time.strptime(value, '%Y-%m-%d')[:3],
                                             **kwargs)
                except ValueError:
                    raise exceptions.ValidationError(self.error_messages['invalid'])

    def pre_save(self, model_instance, add):
        if self.auto_now or (self.auto_now_add and add):
            value = jdatetime.datetime.now()
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super(jDateTimeField, self).pre_save(model_instance, add)

    def get_prep_value(self, value):
        return self.to_python(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        # Casts dates into the format expected by the backend
        if not prepared:
            value = self.get_prep_value(value)

        if isinstance(value, jdatetime.datetime):
            value = value.togregorian()

        if StrictVersion(django.get_version()) >= StrictVersion('1.9'):
            return connection.ops.adapt_datefield_value(value)
        else :
            return connection.ops.value_to_db_datetime(value)



    def value_to_string(self, obj):
        val = self._get_val_from_obj(obj)
        if val is None:
            data = ''
        else:
            d = datetime_safe.new_datetime(val)
            data = d.strftime('%Y-%m-%d %H:%M:%S')
        return data

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.jDateTimeField}
        defaults.update(kwargs)
        return super(jDateTimeField, self).formfield(**defaults)



from django_jalali.forms.widgets import jDateInput, jDateTimeInput
from django import forms
import time
import datetime
import jdatetime
from django.core import validators, exceptions
from django.utils import datetime_safe, formats
from django.utils.translation import ugettext as _
from .widgets import jDateInput, jDateTimeInput

class jDateField(forms.Field):
    widget = jDateInput
    default_error_messages = {
        'invalid': _(u'Enter a valid date.'),
    }

    def __init__(self, input_formats=None, *args, **kwargs):
        super(jDateField, self).__init__(*args, **kwargs)
        self.input_formats = input_formats

    def to_python(self, value):
        """
        Validates that the input can be converted to a date. Returns a Python
        datetime.date object.
        """
        if value in validators.EMPTY_VALUES:
            return None
        if isinstance(value, jdatetime.datetime):
            return value.date()
        if isinstance(value, jdatetime.date):
            return value
        for format in self.input_formats or formats.get_format('DATE_INPUT_FORMATS'):
            try:
                return jdatetime.date(*time.strptime(value, format)[:3])
            except ValueError:
                continue
        raise exceptions.ValidationError(self.error_messages['invalid'])


class jDateTimeField(forms.Field):
    widget = jDateTimeInput
    default_error_messages = {
        'invalid': _(u'Enter a valid date/time.'),
    }

    def __init__(self, input_formats=None, *args, **kwargs):
        super(jDateTimeField, self).__init__(*args, **kwargs)
        self.input_formats = input_formats

    def to_python(self, value):
        """
        Validates that the input can be converted to a datetime. Returns a
        Python datetime.datetime object.
        """
        if value in validators.EMPTY_VALUES:
            return None
        if isinstance(value, jdatetime.datetime):
            return value
        if isinstance(value, jdatetime.date):
            return jdatetime.datetime(value.year, value.month, value.day)
        if isinstance(value, list):
            # Input comes from a SplitDateTimeWidget, for example. So, it's two
            # components: date and time.
            if len(value) != 2:
                raise ValidationError(self.error_messages['invalid'])
            if value[0] in validators.EMPTY_VALUES and value[1] in validators.EMPTY_VALUES:
                return None
            value = '%s %s' % tuple(value)
        for format in self.input_formats or formats.get_format('DATETIME_INPUT_FORMATS'):
            try:
                return jdatetime.datetime(*time.strptime(value, format)[:6])
            except ValueError:
                continue
        raise ValidationError(self.error_messages['invalid'])




from django_jalali import forms as jforms
from django.utils.translation import ugettext as _
from django import forms
from django.contrib.admin.widgets import AdminTimeWidget
from django.contrib.admin.templatetags.admin_static import static
from django.utils.safestring import mark_safe

class AdminjDateWidget(jforms.jDateInput):
    @property
    def media(self):
        js = ["admin/jalalijscalendar/jalali.js", "admin/jalalijscalendar/calendar.js",
              "admin/jalalijscalendar/calendar-setup.js",
              "admin/jalalijscalendar/calendar-fa.js", "admin/jDateTimeShortcuts.js"]
        # css = ["calendar-green.css"]
        return forms.Media(js=[static("admin/js/%s" % path) for path in js],
                           css={'all': ('admin/css/calendar-green.css',)})

    def __init__(self, attrs=None, format=None):
        final_attrs = {'class': 'vjDateField', 'size': '10'}
        if attrs is not None:
            final_attrs.update(attrs)
        super(AdminjDateWidget, self).__init__(attrs=final_attrs, format=format)



class AdminSplitjDateTime(forms.SplitDateTimeWidget):
    """
    A SplitDateTime Widget that has some admin-specific styling.
    """
    def __init__(self, attrs=None):
        widgets = [AdminjDateWidget, AdminTimeWidget]
        # Note that we're calling MultiWidget, not SplitDateTimeWidget, because
        # we want to define widgets.
        forms.MultiWidget.__init__(self, widgets, attrs)

    def format_output(self, rendered_widgets):
        return mark_safe(u'<p class="datetime">%s %s<br />%s %s</p>' % \
            (_('Date:'), rendered_widgets[0], _('Time:'), rendered_widgets[1]))



from django.db import models
from django_jalali.db import models as jmodels

class jDateModel(models.Model):
    objects = jmodels.jManager()
    date    = jmodels.jDateField()
    def __repr__(self):
        return str(self.date)

class jDateModelAutoNow(models.Model):
    objects = jmodels.jManager()
    date    = jmodels.jDateField(auto_now=True)
    def __repr__(self):
        return str(self.date)


class jDateTimeModel(models.Model):
    objects   = jmodels.jManager()
    date_time = jmodels.jDateTimeField()

class jDateTimeModelAutoNow(models.Model):
    objects   = jmodels.jManager()
    date_time = jmodels.jDateTimeField(auto_now=True)



from django.test import TestCase
from django.core import exceptions
from django.db.utils import IntegrityError
import jdatetime
import datetime


from django_jalali.tests.models import jDateModel, jDateModelAutoNow, jDateTimeModel, jDateTimeModelAutoNow

from  custom_settings import change_settings
settings_manager = change_settings(append_apps=['django_jalali.tests'])




class JalaliDateModelTestCase(TestCase):
    fixtures = ['initial_unittest_info.json']
    def tearDown(self):
        global settings_manager
        #settings_manager.revert()

    def test_simple_save(self):
        """test simple function, including saving and retrieving jalali date from database"""
        jd_model = jDateModel(date=jdatetime.date.today())
        jd_model.save()
        d = jDateModel.objects.filter(date=jdatetime.date.today())
        self.assertEqual(d[0].date.togregorian(), datetime.date.today())
        
        jd_model_auto_now = jDateModelAutoNow()
        jd_model_auto_now.save()
        self.assertEqual(jd_model_auto_now.date, jdatetime.date.today())

        j_date_time = jdatetime.datetime.today()
        jdt_model = jDateTimeModel(date_time=j_date_time)
        jdt_model.save()
        dt = jDateTimeModel.objects.filter(date_time=j_date_time)
        self.assertEqual(dt[0].date_time.togregorian(), j_date_time.togregorian())

        j_date_time_auto_now = jDateTimeModelAutoNow()
        j_date_time_auto_now.save()
        self.assertEqual(j_date_time_auto_now.date_time.date(), jdatetime.date.today())
        
        

    def test_diffrent_saving(self):
        jd_model = jDateModel(date='1391-01-01')
        jd_model.save()

        jd_model = jDateModel(date=datetime.datetime.now())
        jd_model.save()

        #jd_model = jDateModel()
        #jd_model.save()

        jd_model = jDateModel(date=jdatetime.datetime.now())
        jd_model.save()

        #to_python
        jdt_model = jDateTimeModel(date_time='1391-01-01')
        #TODO: Find the correct exceptions
        #with self.assertRaises(Exception):
        #    jdt_model.date_time = None

        jdt_model.date_time = datetime.datetime(1390,01,02)
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=datetime.datetime(1390,01,02))[0].date_time , jdt_model.date_time)

        jdt_model.date_time = jdatetime.date.today()
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=jdatetime.date.today())[0].date_time, jdt_model.date_time)

        jdt_model.date_time = datetime.date.today()
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=datetime.date.today())[0].date_time, jdt_model.date_time)

        jdt_model.date_time = datetime.date(1390,01,01)
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=jdatetime.date(1390,01,01))[0].date_time, jdt_model.date_time)

        jdt_model.date_time = '1391-01-01 12:12'
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=jdatetime.datetime(1391,01,01,12,12))[0].date_time, jdt_model.date_time)


        jdt_model.date_time = '1391-01-01 12:12:12'
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time=jdatetime.datetime(1391,01,01,12,12,12))[0].date_time, jdt_model.date_time)

        jdt_model.date_time = '2012-01-01 12:12'
        jdt_model.save()
        #TODO: fixme, I'm broken ! I dunno if I ever should be exist
        #self.assertEqual(jDateTimeModel.objects.filter(date_time=datetime.datetime(2012, 01, 01, 12, 12))[0].date_time, jdt_model.date_time)
        self.assertEqual(jDateTimeModel.objects.filter(date_time='2012-01-01 12:12')[0].date_time, jdatetime.datetime(1390, 10, 11, 12, 12))

        jdt_model.date_time = '2012-01-01'
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time='2012-01-01')[0].date_time, jdatetime.datetime(1390, 10, 11))

        jdt_model.date_time = '2012-01-01 12:12:12'
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time='2012-01-01 12:12:12')[0].date_time, jdatetime.datetime(1390, 10, 11, 12, 12, 12))
 
        jdt_model.date_time = '1391-01-01 12:12:12.2222'
        jdt_model.save()
        self.assertEqual(jDateTimeModel.objects.filter(date_time = '1391-01-01 12:12:12.2222')[0] , jdt_model)

    def test_filtering(self):
        """test filtering"""
        today = jdatetime.date.today()
        jdm_model = jDateModel(date=today)
        jdm_model.save()

        #search by year
        jdm_date = jDateModel.objects.filter(date__year=today.year)
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date.year, today.year)

        #search by string
        jdm_date = jDateModel.objects.filter(date="%s-%s-%s"%(today.year, today.month, today.day))
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)

        #save by geregorian date and retrive by jalali date
        g_2009 = datetime.date(2009,9,9)
        jd_model = jDateModel(date=g_2009)
        jd_model.save()

        jdm_date = jDateModel.objects.filter(date=g_2009)
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, jdatetime.date.fromgregorian(date=g_2009))

        jdm_date = jDateModel.objects.filter(date="%s-%s-%s"%(today.year, today.month, today.day))
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)

        jdm_date = jDateModel.objects.filter(date__in =["%s-%s-%s"%(today.year, today.month, today.day)])
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)
        try :
            s= str(jDateModel.objects.filter(date__how =["%s-%s-%s"%(today.year, today.month, today.day)]))
        except :
            self.assertTrue(True)

        #wrong day value
        with self.assertRaises(exceptions.ValidationError):
            jDateModel.objects.filter(date="%s-%s-%s"%(today.year, today.month, 35))

        #invalid format
        with self.assertRaises(exceptions.ValidationError):
            jDateModel.objects.filter(date="%s%s/%s"%(today.year, today.month, 35))

        #invalid search
        with self.assertRaises(ValueError):
            jDateModel.objects.filter(date__month="12")


    def test_admin_gui(self):
        from django.contrib.auth.models import User
        users =   User.objects.all()
        from django.test.client import Client
        c = Client()
        response = c.post('/admin/', {'username': 'milad', 'password': 'milad'})
        response = c.get('/admin/jtest/jtest/')
        #print response



from django.conf import settings
from django.core.management import call_command
from django.db.models import loading

NO_SETTING = ('!', None)

class TestSettingsManager(object):
    """
    A class which can modify some Django settings temporarily for a
    test and then revert them to their original values later.

    Automatically handles resyncing the DB if INSTALLED_APPS is
    modified.

    """
    def __init__(self):
        self._original_settings = {}

    def set(self, **kwargs):
        for k,v in kwargs.iteritems():
            self._original_settings.setdefault(k, getattr(settings, k,
                                                          NO_SETTING))
            setattr(settings, k, v)
        if 'INSTALLED_APPS' in kwargs:
            self.syncdb()

    def syncdb(self):
        loading.cache.loaded = False
        call_command('syncdb', verbosity=1)

    def revert(self):
        for k,v in self._original_settings.iteritems():
            if v == NO_SETTING:
                delattr(settings, k)
            else:
                setattr(settings, k, v)
        if 'INSTALLED_APPS' in self._original_settings:
            self.syncdb()
        self._original_settings = {}

def change_settings(append_apps) : 
    settings_manager = TestSettingsManager()
    s = []
    for st in settings.INSTALLED_APPS:
        s.append(st)
    for app in append_apps :
        s.append(app)
    settings_manager.set(INSTALLED_APPS =  s)
    return settings_manager




from django_jalali.db.models import jDateField, jDateTimeField
from django.contrib.admin import options
from django_jalali.admin import widgets
from django import forms

options.FORMFIELD_FOR_DBFIELD_DEFAULTS[jDateField]     = {'widget': widgets.AdminjDateWidget }
options.FORMFIELD_FOR_DBFIELD_DEFAULTS[jDateTimeField] = {'form_class':  forms.SplitDateTimeField, 'widget': widgets.AdminSplitjDateTime}
#maybe we can use same time
#models.TimeField:       {'widget': widgets.AdminTimeWidget},
#models.DateTimeField: {
#'form_class': forms.SplitDateTimeField,
#'widget': widgets.AdminSplitDateTime
#}




from django.template.base import Library

register = Library()


@register.filter(expects_localtime=True, is_safe=False)
def jformat(value, arg=None):
    """Formats a date or time according to the given format."""
    if value in (None, ''):
        return ''
    if arg is None:
        arg = "%c"
    try:
        # this should be force_text but because jdatetime module didn't handle
        # unicode strings correctly it's not possible to change it at the moment
        arg = str(arg)
        return value.strftime(arg)
    except AttributeError:
        return ''



from setuptools import setup, find_packages

setup(
        name='django-jalali',
        version='2.2',
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        description = ("Jalali Date support for Django model and admin interface"),
        url = 'http://github.com/slashmili/django-jalali',
        download_url = 'http://github.com/slashmili/django-jalali/tarball/master',
        author = 'Milad Rastian',
        author_email = 'eslashmili _at_ gmail.com',
        keywords = "django jalali",
        license='Python Software Foundation License',
        platforms='any',
        install_requires = ["jdatetime>=1.5","django>=1.7"],
        long_description=open('README.md').read()
)
